#!/usr/bin/env python3
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module that implements unittests cases for Glome Class.
"""

import unittest
from cryptography.hazmat.primitives.asymmetric import x25519

import pyglome
from test import test_vectors


class GlomeTestVector:
    """Class that encapsulates needed components for testing Glome Class."""

    def __init__(self, test_vector, truncated_length):
        self.data = test_vector

        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.kb)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kap)
        self.sender_glomes = pyglome.Glome(peer_key, my_key)

        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.ka)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kbp)
        self.receiver_glomes = pyglome.Glome(peer_key, my_key)

        self.truncated_tag_length = truncated_length

        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.kb)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kap)
        self.truncated_sender_glomes = pyglome.Glome(peer_key, my_key,
                                                     self.truncated_tag_length)
        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.ka)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kbp)
        self.truncated_receiver_glomes = pyglome.Glome(
            peer_key, my_key, self.truncated_tag_length)


class GlomeTestBase:
    """Test Class for Glome Class.

    Implements the logic for tests tag and key generation, as well as tag
    checking.
    """

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.test_vector = None

    def test_keys_generation(self):
        test_vector = self.test_vector
        self.assertEqual(
            test_vector.sender_glomes._send_key,
            test_vector.data.sk + test_vector.data.kb + test_vector.data.ka)
        self.assertEqual(
            test_vector.sender_glomes._receive_key,
            test_vector.data.sk + test_vector.data.ka + test_vector.data.kb)
        self.assertEqual(
            test_vector.receiver_glomes._send_key,
            test_vector.data.sk + test_vector.data.ka + test_vector.data.kb)
        self.assertEqual(
            test_vector.receiver_glomes._receive_key,
            test_vector.data.sk + test_vector.data.kb + test_vector.data.ka)

    def test_tag_generation(self):
        test_vector = self.test_vector
        self.assertEqual(
            test_vector.sender_glomes.tag(test_vector.data.msg,
                                          test_vector.data.counter),
            test_vector.data.tag)

    def test_check_raises_exception_when_incorrect(self):
        test_vector = self.test_vector
        with self.assertRaises(pyglome.IncorrectTagError):
            test_vector.sender_glomes.check(tag=bytes([123]),
                                            msg=test_vector.data.msg,
                                            counter=0)
        with self.assertRaises(pyglome.IncorrectTagError):
            test_vector.receiver_glomes.check(tag=bytes([234]),
                                              msg=test_vector.data.msg,
                                              counter=0)
        with self.assertRaises(pyglome.IncorrectTagError):
            test_vector.sender_glomes.check(
                tag=test_vector.data.tag[:test_vector.truncated_tag_length],
                msg=test_vector.data.msg,
                counter=0)
        with self.assertRaises(pyglome.IncorrectTagError):
            test_vector.truncated_receiver_glomes.check(
                tag=test_vector.data.tag[:test_vector.truncated_tag_length] +
                test_vector.data.tag[:test_vector.truncated_tag_length],
                msg=test_vector.data.msg,
                counter=test_vector.data.counter)
        with self.assertRaises(pyglome.IncorrectTagError):
            test_vector.truncated_receiver_glomes.check(
                tag=test_vector.data.tag[:test_vector.truncated_tag_length] +
                test_vector.data.tag[test_vector.truncated_tag_length::-1],
                msg=test_vector.data.msg,
                counter=test_vector.data.counter)

    def test_check_doesnt_raise_exception_when_correct(self):
        test_vector = self.test_vector
        try:
            test_vector.receiver_glomes.check(test_vector.data.tag,
                                              msg=test_vector.data.msg,
                                              counter=test_vector.data.counter)
            test_vector.truncated_receiver_glomes.check(
                test_vector.data.tag[:test_vector.truncated_tag_length],
                msg=test_vector.data.msg,
                counter=test_vector.data.counter)
            test_vector.truncated_receiver_glomes.check(
                test_vector.data.tag[:test_vector.truncated_tag_length + 2],
                msg=test_vector.data.msg,
                counter=test_vector.data.counter)
        except pyglome.IncorrectTagError:
            self.fail('check() raised IncorrectTagError unexpectedly!')


class GlomeTest1(unittest.TestCase, GlomeTestBase):
    """TestCase based on test vector #1 from protocol documentation"""

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.test_vector = GlomeTestVector(test_vectors.TEST1, 8)


class GlomeTest2(unittest.TestCase, GlomeTestBase):
    """TestCase based on test vector #1 from protocol documentation"""

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.test_vector = GlomeTestVector(test_vectors.TEST2, 8)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PyGLOME is a Python library that provides an API for GLOME protocol.

Basic Usage:
In order for Alice and Bob to communicate, the first step would be to generate
some new keys:

>>> import pyglome
>>> alice_keys = pyglome.generate_keys()
>>> bob_keys = pyglome.generate_keys()

Suppose that Alice knows Bob's `public_key` and wants to send Bob the message
`msg` and no other message have been shared before. Alice will need to:

>>> glome = pyglome.Glome(bob_keys.public, alice_keys.private)
>>> first_tag = glome.tag(msg, counter=0)

And Alice will send Bob both msg, first_tag as well as Alice's public key. On
Bob ends he will need to do the following:

>>> glome = pyglome.Glome(alice_keys.public, bob_keys.private)
>>> try:
...     first_tag = glome.check(first_tag, msg, counter=0)
... except pyglome.TagCheckError as tag_error:
...     ## Handle the exception.
>>> ## do what you have to do
"""

# Bring glome module to top level
from pyglome.glome import (Glome, TagCheckError, IncorrectTagError,
                           TagGenerationError, generate_keys, AutoGlome)



#!/usr/bin/env python3
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module that implements unittests cases for AutoGlome Class.
"""
import unittest
from test import test_vectors
from cryptography.hazmat.primitives.asymmetric import x25519

import pyglome


class AutoGlomeTestVector:
    """Class that encapsulate needed components for testing AutoGlome Class."""

    def __init__(self, test_vector, min_peer_tag_len, skippable_range):
        self.data = test_vector
        self.skippable_range = skippable_range

        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.kb)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kap)
        self.sender_glome = pyglome.AutoGlome(peer_key,
                                              my_key,
                                              min_peer_tag_len=min_peer_tag_len,
                                              skippable_range=skippable_range)

        peer_key = x25519.X25519PublicKey.from_public_bytes(self.data.ka)
        my_key = x25519.X25519PrivateKey.from_private_bytes(self.data.kbp)
        self.receiver_glome = pyglome.AutoGlome(
            peer_key,
            my_key,
            min_peer_tag_len=min_peer_tag_len,
            skippable_range=skippable_range)


class AutoGlomeTestBase:
    """Test keys constructions, tag generation and tag checking for AutoGlome."""

    def __init__(self):
        self.test_vector = None

    def test_check_counters_raise_exceptions_when_incorrect(self):
        test_vector = self.test_vector
        with self.assertRaises(ValueError):
            test_vector.sender_glome.sending_counter = -1
        with self.assertRaises(ValueError):
            test_vector.receiver_glome.sending_counter = 256
        with self.assertRaises(ValueError):
            test_vector.sender_glome.receiving_counter = 280
        with self.assertRaises(ValueError):
            test_vector.receiver_glome.receiving_counter = 280

    def test_check_counters_dont_raise_exceptions_when_correct(self):
        test_vector = self.test_vector
        try:
            test_vector.sender_glome.sending_counter = 0
            test_vector.receiver_glome.sending_counter = 23
            test_vector.sender_glome.receiving_counter = 123
            test_vector.receiver_glome.receiving_counter = 255
        except ValueError:
            self.fail('properties raised ValueError unexpectedly!')

    def test_check_counters_are_correctly_set(self):
        test_vector = self.test_vector
        test_vector.sender_glome.sending_counter = 0
        self.assertEqual(test_vector.sender_glome.sending_counter, 0)
        test_vector.receiver_glome.sending_counter = 23
        self.assertEqual(test_vector.receiver_glome.sending_counter, 23)
        test_vector.sender_glome.receiving_counter = 123
        self.assertEqual(test_vector.sender_glome.receiving_counter, 123)
        test_vector.receiver_glome.receiving_counter = 255
        self.assertEqual(test_vector.receiver_glome.receiving_counter, 255)

    def test_tag(self):
        test_vector = self.test_vector
        test_vector.sender_glome.sending_counter = test_vector.data.counter
        self.assertEqual(test_vector.sender_glome.tag(test_vector.data.msg),
                         test_vector.data.tag)

    def test_skippable_range(self):
        test_vector = self.test_vector
        try:
            test_vector.receiver_glome.receiving_counter = (
                test_vector.data.counter - test_vector.skippable_range) % 256
            test_vector.receiver_glome.check(test_vector.data.tag,
                                             msg=test_vector.data.msg)
            self.assertEqual((test_vector.data.counter + 1) % 256,
                             test_vector.receiver_glome.receiving_counter)
        except pyglome.IncorrectTagError:
            self.fail('check() raised IncorrectTagError unexpectedly!')


class AutoGlomeTest1(unittest.TestCase, AutoGlomeTestBase):
    """Autoglome test using test vector #1 from the protocol documentation."""

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.test_vector = AutoGlomeTestVector(test_vectors.TEST1,
                                               min_peer_tag_len=32,
                                               skippable_range=0)


class AutoTest2(unittest.TestCase, AutoGlomeTestBase):
    """Autoglome test using test vector #2 from the protocol documentation."""

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.test_vector = AutoGlomeTestVector(test_vectors.TEST2,
                                               min_peer_tag_len=8,
                                               skippable_range=10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module that implements unittests cases for Glome Class.
"""

import unittest
import sys

import test.glome_test, test.autoglome_test, test.fuzzing_test


def suite():
    """Suite of test to run"""
    glome_tests = unittest.TestLoader().loadTestsFromModule(test.glome_test)
    autoglome_tests = unittest.TestLoader().loadTestsFromModule(
        test.autoglome_test)
    fuzzing_tests = unittest.TestLoader().loadTestsFromModule(test.fuzzing_test)

    return unittest.TestSuite([glome_tests, autoglome_tests, fuzzing_tests])


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())  # Nice verbosy output

    result = unittest.TestResult()
    suite().run(result)
    sys.exit(len(result.errors) + len(result.failures))  # Correct exitcode



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyglome",
    version="0.0.2",
    author="Google LLC",
    description="A Python implementation of the GLOME protocol",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/google/glome",
    packages=["pyglome"],
    install_requires=[
        "cryptography",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires='>=3.6',
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python GLOME fuzz main.

This module that implement some easy fuzz testing for pyglome. 
"""

import hypothesis
import unittest
from cryptography.hazmat.primitives.asymmetric import x25519

import pyglome


def _glome(private_bytes, public_bytes, msg, tag, counter, min_tag_len):
    """Calls basic utilities of glome class, accepts only documented exceptions.

    The intention of this function is to be fuzzed over to find cases that throw
    unexpected exceptions."""
    try:
        private_key = x25519.X25519PrivateKey.from_private_bytes(private_bytes)
        peer_key = x25519.X25519PublicKey.from_public_bytes(public_bytes)
    except (ValueError, pyglome.ExchangeError):
        return

    try:
        tag_manager = pyglome.Glome(peer_key, private_key, min_tag_len)
    except ValueError:
        return

    # Call to property method to test side effects
    tag_manager.user_keys
    tag_manager.peer_key

    try:
        tag_manager.tag(msg, counter)
    except pyglome.TagGenerationError:
        pass

    try:
        tag_manager.check(tag, msg, counter)
    except (pyglome.IncorrectTagError, pyglome.TagCheckError):
        pass


def _autoglome(private_bytes, public_bytes, msg, tag, counter, min_tag_len,
               skippable_range):
    """Calls basic utilities of autoglome class, accepts only documented exceptions.

    The intention of this function is to be fuzzed over to find cases that throw
    unexpected exceptions."""
    try:
        private_key = x25519.X25519PrivateKey.from_private_bytes(private_bytes)
        peer_key = x25519.X25519PublicKey.from_public_bytes(public_bytes)
    except (ValueError, pyglome.ExchangeError):
        return

    try:
        tag_manager = pyglome.AutoGlome(peer_key,
                                        private_key,
                                        min_peer_tag_len=min_tag_len,
                                        skippable_range=skippable_range)
    except ValueError:
        return

    # Call to property method to test side effects
    tag_manager.user_keys
    tag_manager.peer_key
    tag_manager.sending_counter
    tag_manager.receiving_counter

    try:
        tag_manager.sending_counter = counter
    except ValueError:
        pass

    try:
        tag_manager.receiving_counter = counter
    except ValueError:
        pass

    try:
        tag_manager.tag(msg)
    except pyglome.TagGenerationError:
        pass

    try:
        tag_manager.check(tag, msg)
    except (pyglome.IncorrectTagError, pyglome.TagCheckError):
        pass


@hypothesis.settings(max_examples=10**7)
@hypothesis.given(
    hypothesis.strategies.binary(min_size=32, max_size=32),  #private_bytes
    hypothesis.strategies.binary(min_size=32, max_size=32),  #public_bytes
    hypothesis.strategies.binary(),  #msg
    hypothesis.strategies.binary(min_size=32, max_size=32),  #tag
    hypothesis.strategies.integers(),  #counter
    hypothesis.strategies.integers())  #min_tag_len
def glome_test(private_bytes, public_bytes, msg, tag, counter, min_tag_len):
    """Add hypothesis decorator to _glome function"""
    _glome(private_bytes, public_bytes, msg, tag, counter, min_tag_len)


@hypothesis.settings(max_examples=10**7)
@hypothesis.given(
    hypothesis.strategies.binary(min_size=32, max_size=32),  #private_bytes
    hypothesis.strategies.binary(min_size=32, max_size=32),  #public_bytes
    hypothesis.strategies.binary(),  #msg
    hypothesis.strategies.binary(min_size=32, max_size=32),  #tag
    hypothesis.strategies.integers(),  #counter
    hypothesis.strategies.integers(),  #min_tag_len
    hypothesis.strategies.integers())  #skippable
def autoglome_test(private_bytes, public_bytes, msg, tag, counter, min_tag_len,
                   skippable):
    """Add hypothesis decorator to _autoglome function"""
    _autoglome(private_bytes, public_bytes, msg, tag, counter, min_tag_len,
               skippable)


@hypothesis.settings(max_examples=10**5)
@hypothesis.given(
    hypothesis.strategies.binary(),  #private_bytes
    hypothesis.strategies.binary(),  #public_bytes
    hypothesis.strategies.binary(),  #msg
    hypothesis.strategies.binary(min_size=32, max_size=32),  #tag
    hypothesis.strategies.integers(),  #counter
    hypothesis.strategies.integers())  #min_tag_len
def glome_unsized_keys_test(private_bytes, public_bytes, msg, tag, counter,
                            min_tag_len):
    """Add hypothesis decorator to _glome function"""
    _glome(private_bytes, public_bytes, msg, tag, counter, min_tag_len)


@hypothesis.settings(max_examples=10**5)
@hypothesis.given(
    hypothesis.strategies.binary(),  #private_bytes
    hypothesis.strategies.binary(),  #public_bytes
    hypothesis.strategies.binary(),  #msg
    hypothesis.strategies.binary(min_size=32, max_size=32),  #tag
    hypothesis.strategies.integers(),  #counter
    hypothesis.strategies.integers(),  #min_tag_len
    hypothesis.strategies.integers())  #skippable
def autoglome_unsized_keys_test(private_bytes, public_bytes, msg, tag, counter,
                                min_tag_len, skippable):
    """Add hypothesis decorator to _autoglome function"""
    _autoglome(private_bytes, public_bytes, msg, tag, counter, min_tag_len,
               skippable)


class GlomeTest1(unittest.TestCase):
    """Test Class that check one iteration of each function.

    Uses sample input to test whether the fuzzing function raise unexpected
    exceptions."""

    def __init__(self, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        constant_one = b'\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11\x11'
        self.private_bytes = constant_one
        self.public_bytes = constant_one
        self.msg = b'\x11'
        self.tag = constant_one
        self.counter = 1
        self.min_tag_len = 1
        self.skippable = 1

    def test_glome_fuzz(self):
        """Test glome fuzzing function with trivial example"""
        try:
            _glome(self.private_bytes, self.public_bytes, self.msg, self.tag,
                   self.counter, self.min_tag_len)
        except:
            self.fail('Unexpected exception was raised.')

    def test_autoglome_fuzz(self):
        """Test autoglome fuzzing function with trivial example"""
        try:
            _autoglome(self.private_bytes, self.public_bytes, self.msg,
                       self.tag, self.counter, self.min_tag_len, self.skippable)
        except:
            self.fail('Unexpected exception was raised.')


if __name__ == "__main__":
    glome_test()
    autoglome_test()
    glome_unsized_keys_test()
    autoglome_unsized_keys_test()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module includes test vectors from the protocol reference.
"""


class TestVector:
    """Class that encapsulate needed components for testing.

    Consider a use case where an user A sends a message to user B.

    Attributes:
      kap: A's private key.
      ka: A's public key.
      kbp: B's private key.
      kb:  B's public key.
      counter: number of messages already shared.
      msg: message to share.
      sk: shared secret betweens A and B.
      tag: tag that matches ka, kb, counter and msg.
    """

    def __init__(self, kap: str, ka: str, kbp: str, kb: str, counter: int,
                 msg: str, sk: str, tag: str):
        """Constructor for TestVector Class."""
        self.kap = bytes.fromhex(kap)
        self.ka = bytes.fromhex(ka)
        self.kbp = bytes.fromhex(kbp)
        self.kb = bytes.fromhex(kb)
        self.counter = counter
        self.msg = msg.encode(encoding="ascii")
        self.sk = bytes.fromhex(sk)
        self.tag = bytes.fromhex(tag)


TEST1 = TestVector(
    kap='77076d0a7318a57d3c16c17251b26645df4c2f87ebc0992ab177fba51db92c2a',
    ka='8520f0098930a754748b7ddcb43ef75a0dbf3a0d26381af4eba4a98eaa9b4e6a',
    kbp='5dab087e624a8a4b79e17f8b83800ee66f3bb1292618b6fd1c2f8b27ff88e0eb',
    kb='de9edb7d7b7dc1b4d35b61c2ece435373f8343c85b78674dadfc7e146f882b4f',
    counter=0,
    msg='The quick brown fox',
    sk='4a5d9d5ba4ce2de1728e3bf480350f25e07e21c947d19e3376f09b3c1e161742',
    tag='9c44389f462d35d0672faf73a5e118f8b9f5c340bbe8d340e2b947c205ea4fa3')

TEST2 = TestVector(
    kap='b105f00db105f00db105f00db105f00db105f00db105f00db105f00db105f00d',
    ka='d1b6941bba120bcd131f335da15778d9c68dadd398ae61cf8e7d94484ee65647',
    kbp='fee1deadfee1deadfee1deadfee1deadfee1deadfee1deadfee1deadfee1dead',
    kb='872f435bb8b89d0e3ad62aa2e511074ee195e1c39ef6a88001418be656e3c376',
    counter=100,
    msg='The quick brown fox',
    sk='4b1ee05fcd2ae53ebe4c9ec94915cb057109389a2aa415f26986bddebf379d67',
    tag='06476f1f314b06c7f96e5dc62b2308268cbdb6140aefeeb55940731863032277')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python GLOME library.

This module contains the Glome class and generate_keys function.

Example use: Sender

>>> import pyglome
>>> tag_manager = pyglome.Glome(peer_key)
>>> first_tag = tag_manager.tag(first_msg, 0) # 0 as it is the first msg
>>> second_tag = tag_manager.tag(second_msg, 1)

Example use: Receiver

>>> import pyglome
>>> tag_manager = pyglome.Glome(peer_key, my_private_key)
>>> ## Need to have a private key (paired to the public key
>>> ## that the sender use)
>>> try:
...     tag_manager.check(tag, msg, counter=0):
>>> except pyglome.IncorrectTagError as wte:
...     ## Handle the exception
>>> ## do what you have to do
"""

import os
import hashlib
import hmac
from typing import NamedTuple

from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives import serialization


class KeyPair(NamedTuple):
    """
    NamedTuple-Class that stores a private/public key pair.

    Attributes:
        - private: A private key.
        - public: A public key paired with the private one.
    """
    private: x25519.X25519PrivateKey
    public: x25519.X25519PublicKey


class Error(Exception):
    """Error super-class for any error that is thrown in PyGLOME."""


class TagCheckError(Error):
    """Raised whenever a tag is not correct or the method failed to check it."""


class IncorrectTagError(Error):
    """Raised whenever the tag provided does not match the message and counter."""


class TagGenerationError(Error):
    """Raised whenever a tag could not be generated."""


class ExchangeError(Error):
    """Raised whenever the x25519 key exchange fails."""


def _public_key_encode(public_key: x25519.X25519PublicKey):
    return public_key.public_bytes(serialization.Encoding.Raw,
                                   serialization.PublicFormat.Raw)


def _tag(msg: bytes, counter: int, key: bytes) -> bytes:
    if not 0 <= counter <= 255:
        raise ValueError(f'tag counter (={counter}) must be within [0, 255]')

    message = bytes([counter]) + msg  # msg: N_x|M_n
    digester = hmac.new(key=key, msg=message, digestmod=hashlib.sha256)
    return digester.digest()


class Glome:
    """Implement tag managing functionalities for GLOME protocol.

    This class is initialized by providing your peer's public key and
    optionally your private key. If a private key is not provided, one is
    automatically generated making use of `generate_keys`. Provides methods
    tag (to generate new tags) and check (to check receiving tags).
    """

    MAX_TAG_LEN = 32  # 32 is maximum tag length
    MIN_TAG_LEN = 1

    def __init__(self,
                 peer_key: x25519.X25519PublicKey,
                 my_private_key: x25519.X25519PrivateKey = None,
                 min_peer_tag_len: int = MAX_TAG_LEN):
        """Initialize Glome class.

        Performs the handshake and generates keys.

        Args:
            peer_key: Your peer's public key.
            my_private_key: Your private key.
            min_peer_tag_len: Desired length (in bytes) for the tag.
              Must be an integer in range 1-32.
        Raises:
            ValueError: Raised whenever min_peer_tag_len is not in range 1-32.
            ExchangeError: Raised whenever null shared secret is derived from
              user/peer key pair.
        """

        if my_private_key is None:
            my_private_key, my_public_key = generate_keys()
        else:
            my_public_key = my_private_key.public_key()

        if not Glome.MIN_TAG_LEN <= min_peer_tag_len <= Glome.MAX_TAG_LEN:
            raise ValueError(
                f'min_peer_tag_len (={min_peer_tag_len}) is not within '
                f'[{Glome.MIN_TAG_LEN}, {Glome.MAX_TAG_LEN}]')

        try:
            shared_secret = my_private_key.exchange(peer_key)
        except ValueError as value_error:
            raise ExchangeError(
                'Failed to deduce shared secret') from value_error

        self._send_key = shared_secret + _public_key_encode(
            peer_key) + _public_key_encode(my_public_key)
        self._receive_key = shared_secret + _public_key_encode(
            my_public_key) + _public_key_encode(peer_key)
        self._peer_key = peer_key
        self._my_keys = KeyPair(my_private_key, my_public_key)
        self._min_peer_tag_len = min_peer_tag_len

    @property
    def user_keys(self) -> KeyPair:
        """User's private and public keys used in handshake."""
        return self._my_keys

    @property
    def peer_key(self) -> x25519.X25519PublicKey:
        """Peer's public key used in handshake."""
        return self._peer_key

    def tag(self, msg: bytes, counter: int) -> bytes:
        """Generates a tag from a message and a counter.

        Generates a tag matching some provided message and counter.
        This tag is generated following GLOME protocol specification
        in the context of a communication from the users to theirs peers.

        Args:
           msg: Message to be transmitted.
           counter: Numbers of messages transmitted previously in the
             conversation in this direction (i.e. from the user
             to the peer). Must be an integer in {0,...,255}.
        Returns:
           tag: Tag matching counter and msg.
        Raises:
           TagGenerationError: Raised whenever the method failed to generate tag
             due to ValueError in the arguments.
        """
        try:
            return _tag(msg, counter, self._send_key)
        except ValueError as value_error:
            raise TagGenerationError('Failed to generate tag') from value_error

    def check(self, tag: bytes, msg: bytes, counter: int):
        """Check whether a tag is correct for some message and counter.

        Checks if a tag matches some provided message and counter.
        The method generates the matching tag following GLOME protocol
        specification in the context of a communication from the users'
        peers to the users and then is compared with the tag provided.

        Args:
           tag: Object with the generated tag.
           msg: Object containing received message.
           counter: Numbers of messages transmitted previously in the
             conversation in this direction (i.e. from the peer
             to the user).
        Returns:
           None.
        Raises:
           TagCheckError: Raised whenever the method fails to check the tag
             due to a ValueError in the arguments.
           IncorrectTagError: Raised whenever the tag is incorrect.
        """
        prefix_length = max(len(tag), self._min_peer_tag_len)
        prefix_length = min(prefix_length, Glome.MAX_TAG_LEN)

        try:
            correct_tag = _tag(msg, counter, self._receive_key)[:prefix_length]
        except ValueError as value_error:
            raise TagCheckError('Failed to check the tag') from value_error

        if not hmac.compare_digest(tag, correct_tag):
            raise IncorrectTagError('Tag provided does not match correct tag')


def generate_keys() -> KeyPair:
    """Generates a private/public key pair.

    Provides a random key pair based output of os.urandom. The format
    matches the one requested by Glome Class.

    Args:
       None
    Returns:
       A KeyPair, containing a random private key and the public key derived
       from the generated private key
    """
    private = x25519.X25519PrivateKey.from_private_bytes(
        os.urandom(Glome.MAX_TAG_LEN))
    return KeyPair(private, private.public_key())


class AutoGlome:
    """Adds counter managing functionalities for GLOME protocol.

    This class is initialized by providing your peer's public key and
    optionally your private key. If a private key is not provided, one is
    automatically generated making use of `generate_keys`. On initialization,
    two counter (sending and receiving) are created and set to 0. Provides
    methods tag (to generate new tags) and check (to check receiving tags).
    """

    def __init__(self,
                 peer_key: x25519.X25519PublicKey,
                 my_private_key: x25519.X25519PrivateKey = None,
                 *,
                 min_peer_tag_len: int = Glome.MAX_TAG_LEN,
                 skippable_range: int = 0):
        """Initialize AutoGlome class.

        Performs the handshake, generates keys and counters.
        Args:
           peer_key: Your peer's public key.
           my_private_key: Your private key.
           min_peer_tag_len: Desired length (in bytes) for the tag.
             Must be an integer in range 1-32. keyword only.
           skippable_range: Number of messages that can be missed. keyword only.
             Must be non-negative. For more information please go to check method's
             documentation.
        Raises:
           ValueError: Raised whenever min_peer_tag_len is not in range 1-32 or
             skippable_length is a negative integer.
           ExchangeError: Raised whenever null shared secret is derived from
              user/peer key pair.
        """
        if skippable_range < 0:
            raise ValueError(
                f'skippable_range (={skippable_range}) must be non-negative')

        self.glome = Glome(peer_key,
                           my_private_key,
                           min_peer_tag_len=min_peer_tag_len)
        self._sending_counter = 0
        self._receiving_counter = 0
        self.skippable_range = skippable_range

    @property
    def sending_counter(self) -> int:
        """Number of tags shared from the user to the peer.

        It is incremented each time a new tag is generated. It is always
        one byte long. When the counter gets past 255 it overflows at 0.

        Setter raises ValueError if provided integer is not in range 0-255.
        """
        return self._sending_counter

    @sending_counter.setter
    def sending_counter(self, value: int):
        if not 0 <= value <= 255:
            raise ValueError('Counter must be in range 0-255')
        self._sending_counter = value

    @property
    def receiving_counter(self) -> int:
        """Number of tags the user receives from the peer.

        It is always one byte long. When the counter gets past 255 it restarts at
        0. Every time a message is successfully checked, the receiving_counter is
        set to the next value after the last successful one. Note that if
        skippable_range is n the counter might be increased by any amount in
        range 1-n+1 after a successful check.

        Setter raises ValueError if provided counter is not in range 0-255.
        """
        return self._receiving_counter

    @receiving_counter.setter
    def receiving_counter(self, value: int):
        if not 0 <= value <= 255:
            raise ValueError('Counter must be in range 0-255')
        self._receiving_counter = value

    @property
    def user_keys(self) -> KeyPair:
        """User's private and public keys used in handshake."""
        return self.glome.user_keys

    @property
    def peer_key(self) -> x25519.X25519PublicKey:
        """Peer's public key used in handshake."""
        return self.glome.peer_key

    def tag(self, msg: bytes) -> bytes:
        """Generates a tag from a message.

        Generates a tag matching some provided message and the internal
        sending counter. This tag is generated following GLOME protocol
        specification in the context of a communication from the users to
        theirs peers.

        Args:
           msg: Message to be transmitted.
        Returns:
           tag: Tag matching counter and msg.
        Raises:
           TagGenerationError: Raised whenever the method failed to generate tag
             due to ValueError in the arguments.
        """
        tag = self.glome.tag(msg, self.sending_counter)
        self._sending_counter = (self._sending_counter + 1) % 256
        return tag

    def check(self, tag: bytes, msg: bytes):
        """Check whether a tag is correct for some message.

        Checks if a tag matches some provided message and internal receiving
        counter. The method generates the matching tag following GLOME protocol
        specification in the context of a communication from the users' peers to
        the users and then is compared with the tag provided. If tag checking if
        not successful, the receiving counter remains unchanged.

        If skippable_range if greater than 0, the method try to check the tag
        against all counters in range [receiving_counter, receiving_counter +
        skippable_range], in order, until one is successful. If no one is successful,
        an exceptions is raised and receiving counter remains unchanged.

        Args:
           tag: Object with the generated tag.
           msg: Object containing received message.
        Returns:
           None.
        Raises:
           IncorrectTagError: Raised whenever the tag is incorrect.
        """
        old_counter = self._receiving_counter
        for _ in range(self.skippable_range + 1):
            try:
                self.glome.check(tag, msg, self.receiving_counter)
                self._receiving_counter = (self._receiving_counter + 1) % 256
                return None
            except IncorrectTagError:
                self._receiving_counter = (self._receiving_counter + 1) % 256

        #If no counter matches.
        self._receiving_counter = old_counter
        raise IncorrectTagError('Tag provided does not match correct tag')
